"""
This module contains the functions necessary to fuel the step decorator
factory function ``create_step_decorator``. This step decorator—``@step``—is
used to supervise the execution of a function in a digital worker.

.. code-block:: python

    from thoughtful.supervisor import supervise, step

    @step("1.1")
    def times_two(integers: list) -> None:
        return integers * 2

    @step("1.2")
    def times_three(integer: int) -> None:
        return integer * 3

    def main():
        num_list = [1, 2, 3, 4, 5]

        num_list = times_two(num_list)

        for num in num_list:
            num = times_three(num)

    if __name__ == '__main__':
        with supervise():
            main()

The decorator produced by the factory does not have any immediate support for
records, nor for setting the step status like the step context does. However,
some helpers have been implemented to simulate this functionality. For the
record IDs, the decorator will look for a function kwarg called
``supervisor_record_id`` and use that as the record ID.

.. code-block:: python

    @step("1.2")
    def times_three(integer: int, supervisor_record_id: str = None) -> None:
        return integer * 3

    num = times_three(3, supervisor_record_id=str(3))

As for the step status, a helper function, ``set_step_status``, has been
implemented to set the status of a step.

.. code-block:: python

    @step("1.1")
    def times_two(integers: list) -> None:
        return integers * 2

    num_list = [1, 2, 3, 4, 5]
    num_list = times_two(num_list)
    set_step_status("1.1", "warning")
"""

from __future__ import annotations

import contextlib
import functools
import json
import logging
import warnings
from typing import Any, Callable, Dict, List, Optional

from thoughtful.supervisor.recorder import Recorder
from thoughtful.supervisor.reporting.record import Record
from thoughtful.supervisor.reporting.report_builder import ReportBuilder
from thoughtful.supervisor.reporting.report_builder import StepReportBuilder
from thoughtful.supervisor.reporting.status import Status
from thoughtful.supervisor.reporting.timer import Timer


def create_step_decorator(
    report_builder: ReportBuilder, recorder: Recorder
) -> Callable:
    """
    A step decorator generator that as input receives a ``ReportBuilder``
    object and a ``Recorder`` object. The returned decorator will use these
    objects to record the execution of the decorated function.

    Note that the return type here is `Any`, because proper type hinting
    for decorators is not yet obvious. PEP 612 specifies a way to type
    hint decorators starting in 3.10, but this project supports >= 3.7. See
    more here: https://stackoverflow.com/a/68290080/2597913.

    Args:
        report_builder (ReportBuilder): The report builder to use to record
            the execution of the decorated function.
        recorder (Recorder): The recorder to use to record the execution of
            the decorated function.

    Returns:
        Any: A decorator to attach to functions.
    """

    # The decorator to be returned as the property
    # This handles the inputs to the decorator itself
    def returned_decorator(*step_id):
        """
        A decorator to mark a function as the implementation of a step in a
        digital worker's manifest.

        To include a `supervisor_record_id`, pass it to the decorated function
        as a kwarg. For example,
        ```
            @step(1)
            def my_func(*args, supervisor_record_id: str):
                do stuff
                ...

            my_func(supervisor_record_id="my_record_id")
        ```
        """
        step_id = _step_id_from_args(step_id)

        # The decorator to grab the function callable itself
        def inner_decorator(fn):
            # And the wrapper around the function to call it with its
            # args and kwargs arguments
            @functools.wraps(fn)
            def wrapper(*fn_args, **fn_kwargs):
                return _run_wrapped_func(
                    fn, step_id, report_builder, recorder, *fn_args, **fn_kwargs
                )

            return wrapper

        return inner_decorator

    return returned_decorator


def _run_wrapped_func(
    fn: Callable,
    step_id: str,
    report_builder: ReportBuilder,
    recorder: Recorder,
    *fn_args: Optional[Any],
    **fn_kwargs: Optional[Any],
) -> Any:
    """
    Runs `fn` with the given args `args` and `kwargs`, times how long it
    takes to run, and records the execution of this function under `step_id`
    in the work report.

    Args:
        fn (Callable): The function to run.
        step_id (str): The ID of the step representing the function to
        execute.
        *args (List[Any]): The input arguments to the function.
        **kwargs (Dict[str, Any]): The input keyword arguments to the function.

    Returns:
    Any: Whatever is returned by executing the function
    """

    """Set up the step"""
    caught_exception: Optional[Exception] = None
    step_final_status: Status

    # Time the function
    func_timer = Timer()
    func_timer.start()

    """Run the step"""
    try:
        fn_result = fn(*fn_args, **fn_kwargs)
        step_final_status = Status.SUCCEEDED
        record_final_status = Status.SUCCEEDED
    except Exception as ex:
        fn_result = None
        caught_exception = ex
        step_final_status = Status.FAILED
        record_final_status = Status.FAILED
    timer_result = func_timer.end()

    """Step teardown"""
    # this checks if a record id was passed to the function and sets record_id
    # as none if it is not included in the kwargs
    record_id: str = fn_kwargs.get("supervisor_record_id")
    record = Record(record_id, record_final_status) if record_id else None

    # Create a new step report
    new_report = StepReportBuilder(
        step_id=step_id,
        start_time=timer_result.start,
        end_time=timer_result.end,
        status=step_final_status,
        message_log=recorder.messages,
        data_log=recorder.data,
        record=record,
    )
    report_builder.workflow.append(new_report)

    # Reset the recorder for the next step
    recorder.messages = []
    recorder.data = []

    """Passthrough the step's return value or raised exception"""
    # If the step failed and raised an exception, raise that instead
    # of returning None
    if caught_exception:
        raise caught_exception

    # Passthrough the function's returned data back up
    return fn_result


def _step_id_from_args(args: tuple) -> str:
    """
    Joins decorator number args into a step ID string. This was more necessary
    when the user was expected to pass in a series of integers as the step ID,
    but now it's just a string. This is still here for backwards compatibility.

    For example,
    ```
        @step(1, 2, 3)
        def my_func():
            ...
    ```
    would return "1.2.3".


    Args:
        args (tuple): The arguments to the decorator.

    Returns:
        str: The step ID.
    """
    if len(args) > 1:
        warnings.warn(
            "Passing multiple step ids to StepContext is deprecated. "
            "Instead, pass a single step id string with dots, ie '1.1'",
            DeprecationWarning,
        )

    return ".".join([str(arg) for arg in args])


def to_safe_jsonable(value: Any) -> Any:
    """
    Returns a version of `value` that can be converted into JSON
    format using the `json` library.

    Args:
        value: The value to ensure is safe to JSON serialize.

    Returns:
        Any: The JSON safe value
    """
    # If the standard JSON library knows how to encode the object,
    # go with that
    with contextlib.suppress(TypeError):
        json.dumps(value)
        return value
    # Otherwise, check if the object has a .__json__() method
    json_attribute = getattr(value, "__json__", None)
    if callable(json_attribute):
        try:
            return value.__json__()
        except Exception:
            # [CX-234] If the call fails, warn and continue trying other options
            logging.warning("__json__ method raised exception", exc_info=True)

    # Use the raw string as a last resort
    return str(value)



# isort: skip_file due to the importance of the python parsing order of these
# modules

from warnings import warn

from thoughtful.supervisor import *  # noqa

warn(
    "Using `import supervisor` is deprecated and will be removed in the future."
    " Please use `import thoughtful.supervisor` instead.",
    DeprecationWarning,
)



"""
This module handles one of the primary functions of Supervisor, which is to
track the timing of steps, as well as the duration of the entire run.
"""

from __future__ import annotations

import datetime
import time
from dataclasses import dataclass
from typing import Optional


@dataclass
class TimerResult:
    """
    The start time, end time, and duration of running a timer.
    """

    start: datetime.datetime
    """
    datetime.datetime: The start time of the timer.
    """

    end: datetime.datetime
    """
    datetime.datetime: The end time of the timer.
    """

    duration: datetime.timedelta
    """
    datetime.timedelta: The duration of the timer.
    """


@dataclass
class Timer:
    """
    Times an action by calling `start()` and `end()`.
    """

    start_time: Optional[datetime.datetime] = None
    """
    datetime.datetime, optional: The start time of the timer.
    """

    end_time: Optional[datetime.datetime] = None
    """
    datetime.datetime, optional: The end time of the timer.
    """

    duration: Optional[datetime.timedelta] = None
    """
    datetime.timedelta, optional: The duration of the timer.
    """

    _perf_counter_start_value: Optional[float] = None

    def start(self) -> datetime.datetime:
        """
        Starts the timer.
        """
        self._perf_counter_start_value = time.perf_counter()
        self.start_time = datetime.datetime.now()
        return self.start_time

    def end(self) -> TimerResult:
        """
        Ends the timer and records the duration.

        Returns:
            TimerResult: How long the timer ran and when it started.
        """
        self.end_time = datetime.datetime.now()
        self.duration = self.end_time - self.start_time

        return TimerResult(
            start=self.start_time, end=self.end_time, duration=self.duration
        )



"""
Defines json payloads to be delivered to the streaming endpoint
"""
import datetime

import thoughtful
from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.reporting.step_report import StepReport
from thoughtful.supervisor.reporting.streaming_actions import Action


class StreamingPayload:
    """Defines core payload data required for all streaming requests"""

    def __init__(self, run_id: str):
        self.run_id = run_id
        self.client = thoughtful.supervisor.__name__
        self.version = thoughtful.__version__

    def __json__(self):
        return {
            "run_id": self.run_id,
            "client": self.client,
            "version": self.version,
        }


class StepReportStreamingPayload(StreamingPayload):
    """
    Defines the payload for a step report action, this occurs when entering or
    exiting a step context
    """

    def __init__(self, step_report: StepReport, run_id: str):
        self.step_report = step_report
        super().__init__(run_id)

    def __json__(self):
        _json = super().__json__()
        _json.update(
            action=Action.STEP_REPORT,
            payload={
                "step_report": self.step_report.__json__(),
                "timestamp": str(datetime.datetime.utcnow().isoformat()),
            },
        )
        return _json


class BotManifestStreamingPayload(StreamingPayload):
    """
    Defines the payload for posting the manifest to the opus streaming endpoint
    occurring when entering supervisor's main context
    """

    def __init__(self, manifest: Manifest, run_id: str):
        self.manifest = manifest
        super().__init__(run_id)

    def __json__(self):
        _json = super().__json__()
        _json.update(
            action=Action.BOT_MANIFEST,
            payload={
                "bot_manifest": self.manifest.__json__(),
                "timestamp": str(datetime.datetime.utcnow().isoformat()),
            },
        )
        return _json



from __future__ import annotations

import logging
from typing import Optional

from aws_requests_auth.aws_auth import AWSRequestsAuth
from requests import Session

from thoughtful.supervisor.manifest import Manifest
from thoughtful.supervisor.reporting.step_report import StepReport

from thoughtful.supervisor.reporting.streaming_payloads import (  # isort:skip
    StepReportStreamingPayload,
    BotManifestStreamingPayload,
    StreamingPayload,
)

logger = logging.getLogger(__name__)


class StreamingCallback(Session):
    def __init__(self, run_id: str, callback_url: str, aws_auth: AWSRequestsAuth):
        super().__init__()
        self.run_id = run_id
        self.callback_url = callback_url
        self.auth = aws_auth

    @classmethod
    def from_env_defaults(
        cls, auth: AWSRequestsAuth, run_id: str, callback_url: str
    ) -> Optional[StreamingCallback]:
        if run_id and callback_url:
            return cls(run_id=run_id, callback_url=callback_url, aws_auth=auth)
        logger.warning("Missing job id or callback url values")

    def post(self, payload: StreamingPayload, **kwargs):
        logger.info(f"Posting to {self.callback_url} {StreamingPayload}")
        try:
            response = super().post(
                self.callback_url, json=payload.__json__(), timeout=10, **kwargs
            )
            if not response.ok:
                logger.warning(
                    f"Invalid response {response.status_code}: {response.text}"
                )
            logger.info(f"Received {response.status_code}: {response.text}")
            return response
        except Exception:
            logger.exception("Could not post step payload to endpoint")

    def post_step_update(self, report: StepReport):
        return self.post(StepReportStreamingPayload(report, self.run_id))

    def post_manifest(self, manifest: Manifest):
        return self.post(BotManifestStreamingPayload(manifest, self.run_id))
